#!/tvbin/tivosh

# endpad.tcl
# (c) Stuart Anderton, 2005
# Version 1.5.4, 18 Jan 2007
#
# Program to add padding at the start and end of every programme if it does not
# cause a clash.
#
# Padding is added a few minutes before it is needed to minimise the possibility
# of clashes being caused by the padding. Runs in the background, 24/7.
#
# This code may be distributed freely so long as this header is not changed
# and no-one gets charged for it. Distribution on disks or ISOs sold on eBay or
# elsewhere is specifically prohibited.
#
# EndPad is entirely free to use, but next time its saves you a verbal beating
# thanks to a caught ending that would otherwise be missed, please consider an
# appropriate donation to Cancer Research through this link:
#
#     http://www.justgiving.com/brigidphillips
#
# Please put "EndPad" in the comment.
#
# INSTALLATION AND EXAMPLE USAGE
# ==============================
#
# Copy endpad.tcl to the TiVo using binary-mode FTP. Then:
#
#   chmod 755 endpad.tcl
#
# It can be run with, eg.,
#
#   ./endpad.tcl -e 10
#
# In this case, EndPad will try to add 10 minutes padding to the end of all
# programmes, but only if this does not cause a clash with another scheduled
# recording.
#
# Stop with:
#
#   ./endpad.tcl -stop
#
# To start EndPad every time the TiVo starts up, add the following line
# to /etc/rc.d/rc.sysinit.author:-
#
#   /var/hack/endpad.tcl -e 10 -auto >> /dev/null &
#
# Run ./endpad -help for information on other options. See the comments in
# endpad.config for information on the configuration file.
#
# CREDITS
# =======
#
# See changes.txt for details of recent changes to EndPad.
#
# EndPad was created by Stuart Anderton
# Thanks to Dibblah and ccwf for suggestions and refinements
# Log rotation code courtesy of LJ
# Subs quality change code by Milhouse
# Config file, channel-specific, time-dependent settings, additional
# documentation, and other changes by maxwells_daemon.
# Title matching, auto reload, negpad integration by tym, using manual recording
# code adapated from manrec.itcl 0.2.0 by Christopher R. Wingert
# Portions adapted from code published on the O'Reilly site from "TiVo Hacks"
# Other bits based on noreddot.tcl by LJ, as well as the TivoWeb code
#
# SUPPORT
# =======
#
# Questions, suggestions, and bug reports to the TiVo UK Forum on
# www.tivocommunity.com:
#
#   http://www.tivocommunity.com/tivo-vb/showthread.php?t=286631

######################################################################
### Usage
######################################################################

proc usage {} {
  global defaultConfigFile tz
  puts "Usage:"
  puts "  [file tail [info script]] -s X -e Y -n W -f Q -sugqual Z -sugeq -config F -tz T -auto -debug -stop"
  puts "All arguments are optional:-"
  puts "  -s X        Start padding X minutes"
  puts "  -e Y        End padding Y minutes"
  puts "  -n W        Negative start padding W minutes"
  puts "  -f Q        Q=0 or 1. If 1, force negative padding even if no channel change"
  puts "  -sugqual Z  Force suggestions to quality Z (0, 40, 75, or 100)"
  puts "  -sugeq      Do not cancel suggestions to pad other suggestions"
  puts "  -config F   Read per-channel settings from file F (default=$defaultConfigFile)"
  puts "  -tz T       Timezone (+/-hours) for config and displayed times (default=[timezoneString $tz])"
  puts "  -auto       Use if run from startup script (otherwise stops previous EndPad)"
  puts "  -debug      Don\'t put in background and print messages to screen"
  puts "  -stop       Stop existing endpad.tcl process"
  puts "  -help       Display this help message"
}

######################################################################
### Settings
######################################################################

# The version number
set epversion "1.5.4"

# Directory EndPad is in
set eppath [file dirname [info script]]

# Default configuration array. Keys are:-
#   startpad: Start padding in minutes
#   endpad:   End padding in minutes
#   sugqual:  The recording quality for Suggested recordings (ie. not scheduled recording)
#             Basic = 0, Medium = 40, High = 75, Best = 100, default = -1
#   sugeq:    Should padding on a suggestion cancel an immediately following suggestion?
array set defopt {startpad 0 endpad 0 sugqual -1 sugeq 1 negpad 0 maxneg -1 forceneg 0}

# All option names (value not used)
array set selopt [array get defopt]
array set selopt {channel 1 validity 1 program 1}

# Flags (options that don't take a value) and the value they set
array set noargval {sugeq 0}

# The name of the file containing the PID of the backgrounded process
set pidfile "/var/run/endpad.pid"

# The name of the file with the kill instruction
set killfile "$eppath/endpad.kill"

# The name of the log file
set logfile "/var/log/endpad.log"
set Ologfile "/var/log/Oendpad.log"

# The name of the summary file
set msgfile "/var/log/endpad.msg"
set Omsgfile "/var/log/Oendpad.msg"

# The output of the script to stderr, in case of sudden failures
set errfile "$eppath/endpad.err"

# Configuration file to look for (can be overridden with -f)
set defaultConfigFile "$eppath/endpad.config"
set configMtime 0

# Default timezone is GMT
set tz 0

set sweekday { Sun    Mon    Tue     Wed       Thu      Fri    Sat      }
set lweekday { Sunday Monday Tuesday Wednesday Thursday Friday Saturday }
set i 0
foreach sd $sweekday ld $lweekday {
  set weekday([string tolower $sd]) $i
  set weekday([string tolower $ld]) $i
  incr i
}

######################################################################
### Procedures
######################################################################

# Check if a process referenced by a PID file is running
proc checkpid pidfile {
  global errfile
  if { [file exists $pidfile] } {
    set f [open $pidfile]
    gets $f runningpid
    close $f
    if { ![file exists "/proc/$runningpid"] } {
      file delete -force $pidfile
      puts "A stale PID file was found and has been deleted."
      puts "This means that this script terminated unexpectedly last time. Loading now."
      puts "The errors logged (if any) before the last termination were:"
      if { [file exists $errfile] } {
        set f [open $errfile]
        gets $f errline
        while { $errline!=""} {
          puts "         $errline"
          gets $f errline
        }
      }
      return 1
    }
    return 0
  }
  return -1
}

# Background error catcher
proc bgerror errmessage {
  global logchan errorInfo
  log $errmessage
  log $errorInfo
  close $logchan
  close $msgchan
}

# make log entry
proc log {entry} {
  global logchan datestamp tz
  puts $logchan "[clock format [expr [clock seconds]+$tz] -format $datestamp] : $entry"
}

# make log and summary entries
proc msg {entry} {
  global logchan msgchan datestamp tz
  set str "[clock format [expr [clock seconds]+$tz] -format $datestamp] : $entry"
  puts  $logchan $str
  flush $logchan
  puts  $msgchan $str
  flush $msgchan
}

# From mfslib.tcl with small amendment, deleting its console output
proc RetryTransaction { body } {
  global errorInfo errorCode
  while { 1 } {
    set code [catch {transaction {uplevel $body}} string]
    if { $code == 0 } {
      return $string
    } elseif { $code == 1 } {
      if { $errorCode == "errTmActiveLockConflict" ||
        $errorCode == "errTmBackgroundHoldoff" ||
        $errorCode == "errFsLockConflict" } {
        # puts "retrying after $errorCode ..."
        after [random 2000 4000]
        # retry the transaction
        continue
      }
      error $string $errorInfo $errorCode
    } elseif { $code == 2 } {
      return -code return $string
    } elseif { $code == 3 } {
      return -code break $string
    } elseif { $code == 4 } {
      return -code continue $string
    }
  }
}

# Print log header
proc logheader {} {
  global suggestionRecQuality sugcancancelsug startpadding endpadding negativepadding versionstring tz forcenegpad
  log " Start padding set to $startpadding seconds"
  log " End padding set to $endpadding seconds"
  log " Negative start padding set to $negativepadding seconds"
  log " Force Negative startpadding is $forcenegpad"
  if {$sugcancancelsug == 1} {
    log " Suggestions will be padded, even if that means cancelling a subsequent suggestion"
  } else {
    log " Suggestions will not be padded if another suggestion immediately follows them"
  }
  if [info exists versionstring] {
    log " Detected TiVo software version $versionstring"
  }
  switch -- $suggestionRecQuality {
    "-1" {log " Suggestion quality change disabled"}
    "0" {log " Suggestions will be recorded in Basic"}
    "40" {log " Suggestions will be recorded in Medium"}
    "75" {log " Suggestions will be recorded in High"}
    "100" {log " Suggestions will be recorded in Best"}
    default  {log " Suggestion quality change - invalid setting"}
  }
  log " Timezone [timezoneString $tz]"
  showConfig
}

# To Delete suggestions!
proc DeleteSuggestion { recfsid } {
  global dbPoolSize
  set db [dbopen $dbPoolSize]
  set canceldate [expr [clock seconds] / 86400]
  set canceltime [expr [clock seconds] % 86400]
  RetryTransaction {
    set rec [db $db openid $recfsid]
    set state [dbobj $rec get State]
    if { $state != 6 } {
      dbclose $db
      return 0
    } else {
      dbobj $rec set CancelReason 10
      dbobj $rec set CancelDate $canceldate
      dbobj $rec set CancelTime $canceltime
      dbobj $rec set ErrorString "Suggestion Deleted by EndPad"
      dbobj $rec set State 2
    }
  }
  dbclose $db
  return 1
}

proc AddNegativeStartPadding { recfsid npad } {
  global dbPoolSize version logchan
  set db [dbopen $dbPoolSize]

  RetryTransaction {
    set nrec [db $db openid $recfsid]
    if {$version >= 3} {
      set nrecbeh [dbobj $nrec get RecordingBehavior]
      set nselectiontype [dbobj $nrecbeh get PresentationBehavior]
      set pgb [dbobj $nrecbeh get ProgramGuideBehavior]
    } else {
      set nselectiontype [dbobj $nrec get SelectionType]
    }
    set nshowing [dbobj $nrec get Showing]

    set nstartdate [dbobj $nshowing get Date]
    set nstarttime [dbobj $nshowing get Time]
    set nduration [dbobj $nshowing get Duration]
    set nprogram [dbobj $nshowing get Program]
    set ntitle [dbobj $nprogram get Title]
    set nexistingstartpadding [dbobj $nrec get StartPadding]
    set nexistingendpadding [dbobj $nrec get EndPadding]
  }
  if {$nexistingstartpadding == ""} {set nexistingstartpadding 0}
  if {$nexistingendpadding == ""} {set nexistingendpadding 0}

  # Actual start, end and duration times

  set starttime [expr $nstarttime - $nexistingstartpadding]
  set endtime [expr $nstarttime + $nduration + $nexistingendpadding]
  set duration [expr $nduration + $nexistingendpadding + $nexistingstartpadding]

  log " Original start time for $ntitle is $starttime"

  # Altered start and duration times

  set startdate $nstartdate
  set starttime [expr $starttime + $npad]
  set duration [expr $duration - $npad]

  if { $duration < 1 } {
    return -1
  }

  # check if the new start time is on the following day!
  if { $starttime >= 86400 } {
    incr startdate 1
    set starttime [expr $starttime - 86400]
  }
  set stopdate $startdate
  set stoptime [expr $starttime + $duration]
  # check if the new stop time is on the following day!
  if { $stoptime >= 86400 } {
    incr stopdate 1
    set stoptime [expr $stoptime - 86400]
  }


  log " New start time is $starttime"

  # Cancel Program requiring negative padding

  set canceldate [expr [clock seconds] / 86400]
  set canceltime [expr [clock seconds] % 86400]
  RetryTransaction {
    set rec [db $db openid $recfsid]

    set programsourcefsid {}
    set showing [dbobj $rec get Showing]
    set station [dbobj $showing get Station]
    set stationfsid [dbobj $station fsid]
    set program [dbobj $showing get Program]
    set programfsid [dbobj $program fsid]
    set et [dbobj $rec get ExpirationTime]
    set ed [dbobj $rec get ExpirationDate]
    set rq [dbobj $rec get RecordQuality]
    set sc [dbobj $rec get Score]
    set br [dbobj $rec get BitRate]
    set st [dbobj $rec get SelectionType]
    if { $st == 9 } {
      foreach programsource [dbobj $rec get ProgramSource] {
        lappend programsourcefsid [dbobj $programsource fsid]
      }
    }
    dbobj $rec set CancelReason 10
    dbobj $rec set CancelDate $canceldate
    dbobj $rec set CancelTime $canceltime
    dbobj $rec set ErrorString "Programme Deleted by EndPad for Negative Padding"
    dbobj $rec set State 2

  }
  log " Cancelled old recording"

  # Recreate by manual recording

  RetryTransaction {
    log "  Creating Recording..."
    flush $logchan

    set recording [ db $db create Recording ]
    set newrecfsid [dbobj $recording fsid]
    dbobj $recording set BitRate $br
    dbobj $recording set ErrorString "Manual Negative Padding (endpad)"
    dbobj $recording set ExpirationDate $ed
    dbobj $recording set ExpirationTime $et

    foreach psfsid $programsourcefsid {
      set ps [db $db openid $psfsid]
      dbobj $recording add ProgramSource $ps
    }

    ###########################################################
    # NOTE: To create an entry that is not flagged as manual, #
    #       use SelectionType=3 and ProgramGuideBehavior=1.   #
    #       default: ProgramGuideBehavior=3.                  #
    #       Not really recommended as it will cause problems  #
    #       while listed in the ToDo list.                    #
    ###########################################################

    # Create a manual(5) entry and place in the todo(6) list
    if {$rq != ""} {
      dbobj $recording set RecordQuality $rq
    }
    dbobj $recording set Score $sc
    dbobj $recording set SelectionType $st
    dbobj $recording set StartDate $startdate
    dbobj $recording set StartTime $starttime
    dbobj $recording set State 6
    dbobj $recording set StopDate $stopdate
    dbobj $recording set StopTime $stoptime

    if {$version >= 3} {
      # Only needed for 3.x
      dbobj $recording set StreamFileSize 0
      dbobj $recording set UsedBy 1

      #
      # Create the RecordingBehavior subobject.
      #
      set recordingbehavior [db $db createsub RecordingBehavior $recording]
      dbobj $recordingbehavior set DiskBehavior 1
      # dbobj $recordingbehavior set PresentationBehavior 2
      dbobj $recordingbehavior set PresentationBehavior $nselectiontype
      dbobj $recordingbehavior set ProgramGuideBehavior $pgb
      # dbobj $recordingbehavior set ProgramGuideBehavior 3
      dbobj $recordingbehavior set TunerBehavior 3
      dbobj $recording set RecordingBehavior $recordingbehavior
    }

    log "   Creating Showing..."
    flush $logchan
    #
    # Create the showing subobject.
    #
    set showing [db $db createsub Showing $recording]
    dbobj $recording set Showing $showing
    dbobj $showing set Date $startdate
    dbobj $showing set Duration $duration
    dbobj $showing set Program [db $db openid $programfsid]
    dbobj $showing set Reason 2
    dbobj $showing set Station [db $db openid $stationfsid]
    dbobj $showing set Time $starttime
    #dbobj $showing set TvRating $ratingstr
    log "  Done."
    flush $logchan
  }

  # Original recording creation code from shanew's negative padding code.
  # Does not work with v3 tivos.
#  RetryTransaction {
#    set recording [db $db create Recording]
#    set newrecfsid [dbobj $recording fsid]
#    dbobj $recording set BitRate $br
#    dbobj $recording set ExpirationTime $et
#    dbobj $recording set ExpirationDate $ed
#    if {$rq != ""} {
#      dbobj $recording set RecordQuality $rq
#    }
#    dbobj $recording set Score $sc
#    if {$version >= 3} {
#      dbobj $recording set StreamFileSize 0
#      dbobj $recording set UsedBy 1
#
#      set recordingbehavior [db $db createsub RecordingBehavior $recording]
#      dbobj $recordingbehavior set DiskBehavior 1
#      dbobj $recordingbehavior set PresentationBehavior $st
#      dbobj $recordingbehavior set ProgramGuideBehavior 3
#      dbobj $recordingbehavior set TunerBehavior 3
#      dbobj $recording set RecordingBehavior $recordingbehavior
#    } else {
#      dbobj $recording set SelectionType $st
#    }
#
#    if { $st == 9 } {
#      dbobj $recording set ProgramSource [db $db openid $programsourcefsid]
#    }
#
#    set showing [db $db createsub Showing $recording]
#    dbobj $showing set Date $startdate
#    dbobj $showing set Time $starttime
#    dbobj $showing set Duration $duration
#    dbobj $showing set Station [db $db openid $stationfsid]
#    dbobj $showing set Program [db $db openid $programfsid]
#    dbobj $recording set Showing $showing
#
#    dbobj $recording set State 6
#    log " Attempting commit"
#  }

  dbclose $db

  return $newrecfsid
}

proc localTime { time {timezone ""} } {
  global tz
  if {$timezone == ""} { set timezone $tz }
  return [expr {($time+$timezone)%86400}]
}

proc gmTime { time {timezone ""} } {
  global tz
  if {$timezone == ""} { set timezone $tz }
  return [expr {($time-$timezone)%86400}]
}

proc timeString { time {timezone ""} } {
  set time [localTime $time $timezone]
  foreach tt {ss mm hh} f {60 60 24} {
    set $tt  [expr {$time % $f}]
    set time [expr {$time / $f}]
  }
  set v [format "%02d:%02d:%02d" $hh $mm $ss]
  regsub {:00$} $v "" v
  return $v
}

proc validityString { validity } {
  global sweekday tz
  if {[lindex $validity 2]} {
    foreach v [lrange $validity 0 1] {
      set v [expr {($v+$tz)%604800}]
      lappend vd $v
      lappend vt [timeString [expr {$v%86400}] 0]
    }
    # For day, use $v from start time
    return "[lindex $sweekday [expr {[lindex $vd 0]/86400}]]/[join $vt "-"]"
  } else {
    return "[timeString [lindex $validity 0]]-[timeString [lindex $validity 1]]"
  }
}

proc timezoneNumber { zone } {
  set neg [regsub {^-} $zone "" newtz]
  set newtz [timeNumber $newtz 0]
  if {$newtz == ""} {
    puts "bad timezone: $zone"
    usage
    exit 0
  }
  return [expr {$neg ? -$newtz : $newtz}]
}

proc timezoneString { zone } {
  if {$zone >= 0} {
    return [timeString $zone 0]
  } else {
    return "-[timeString [expr abs($zone)] 0]"
  }
}

proc timeNumber { time {timezone ""} } {
  foreach tt {hh mm ss} vt [split $time ":"] {
    if {$vt == ""} {
      set $tt 0
    } elseif [regexp {^[0-5]?[0-9]$} $vt] {
      scan $vt {%d} $tt
    } else {
      return
    }
  }
  if {$hh < 24 && $mm < 60 && $ss < 60} {
    return [gmTime [expr {60*((60*$hh) + $mm) + $ss}] $timezone]
  } else {
    return
  }
}

# Returns validity list: {start end weekly}
# if weekly=0, start and end are in seconds since midnight
# if weekly=1, start and end are in seconds since Sunday midnight
proc validityNumbers { validity } {
  global weekday tz
  if [info exists weekday([string tolower $validity])] {
    set day $weekday([string tolower $validity])
    set validity "0-0"
  } elseif [regexp {^(.*)/(.*)} $validity {} day validity] {
    set day [string tolower $day]
    if [info exists weekday($day)] {
      set day $weekday($day)
    } else {
      return
    }
  } else {
    set day ""
  }

  set vs [split $validity "-"]
  if {[llength $vs] != 2} { return }

  if {$day == ""} {
    foreach v $vs {
      set vt [timeNumber $v]
      if {$vt == ""} { return }
      lappend vl $vt
    }
    lappend vl 0

  } else {

    foreach v $vs {
      set vt [timeNumber $v 0]
      if {$vt == ""} { return }
      lappend vw $vt
    }

    # Convert to seconds since Sunday midnight,
    # adding 1 day to end time if it wraps round midnight
    foreach v $vw dt "0 [expr {[lindex $vw 0] >= [lindex $vw 1]}]" {
      lappend vl [expr {(($day+$dt)*86400+$v-$tz)%604800}]
    }
    lappend vl 1
  }
  return $vl
}

# Re/load the configuration
proc reloadConfig {configFile} {
  global config cmdopt defopt
  catch { unset config }
  set config() [list [concat [array get defopt] linenum 0]]
  readConfig $configFile

  # command line arguments override config file settings
  if [array size cmdopt] {
    lappend config() [array get cmdopt]
  }

  set sf 1000000
  foreach chan [array names config] {
    # Define precedence
    set pc {}
    foreach c $config($chan) {
      catch {unset a}
      array set a $c
      set vp 0; set lp 0
      if [info exists a(validity)] {
        if {[lindex $a(validity) 2]} {
          set vp [expr {([lindex $a(validity) 1]-[lindex $a(validity) 0])%86400 == 0 ? 2 : 3}]
        } else {
          set vp 1
        }
      }
      if [info exists a(linenum)] {
        set lp $a(linenum)
        if {$lp < 0 || $lp >= $sf} {
          set lp [expr {$sf-1}]
        }
      }
      lappend c precedence [expr {([info exists a(program)]*4 + $vp)*$sf + $lp}]
      lappend pc $c
    }
    # Sort by precedence, and for efficiency, merge entries with the same
    # validity and program selections.
    # May not merge every case, but should catch most.
    catch {unset l}
    set nc {}
    foreach c [lsort -integer -index end $pc] {
      catch {unset o}
      array set o $c
      if {[info exists l] &&
          ([info exists o(validity)] ? $o(validity) : {}) == ([info exists l(validity)] ? $l(validity) : {}) &&
          ([info exists o(program)]  ? $o(program)  : {}) == ([info exists l(program)]  ? $l(program)  : {})} {
        array set l $c
        set nc [lreplace $nc end end [array get l]]
      } else {
        lappend nc $c
        catch {unset l}
        array set l $c
      }
    }
    set config($chan) $nc
  }
}

# Read configuration file.
# Keywords are from $defopt and $selopt above, ie.
#   timezone hh:mm   Set the timezone (don't use this on ozTiVo)
#   channel N   This line's settings apply to channel N only.
#   validity hh:mm-hh:mm   This line's settings only apply if the programme's
#               start time is within the specified interval. The times can be
#               specified as hh, hh:mm, or hh:mm:ss and may be limited to one
#               day of the week with Day/hh:mm etc. Note that the interval
#               can wrap around midnight.
#   program P   This line's settings apply if the name of the program matches
#               the supplied pattern, P (may be quoted using {} or "").
#   programme P As above, but with British spelling!
#   startpad X  Start padding X minutes
#   endpad Y    End padding Y minutes
#   negpad W    Start the following programme late to allow W minutes endpad
#   maxneg M    Maximum late starting for matching programme is M minutes
#   forceneg Q  If 1, do negative padding even if no channel change is required
#   sugqual Z   The recording quality for suggested recordings
#               Basic = 0, Medium = 40, High = 75, Best = 100, default = -1
#   sugeq       Padding of suggestions on this channel does not cause subsequent
#               suggestions on other channels to be cancelled.
# Keywords can be abbreviated as follows: c(hannel), p(rogram), v(alidity),
# s(tartpad), e(ndpad), n(egpad), m(axneg), tz (timezone)
# Keywords may optionally be preceded by "-". Comments start with "#".
# Examples:-
#   startpad 1 endpad 3                    # global default (same as -s 1 -e 3)
#   channel BBC1STH,BBC2STH,BBC4 endpad 15 # override endpad for the BBC
#   channel C4 validity 18-5 endpad 10     # endpad for 6pm to 5am on Channel 4
#   programme Buffy endpad 20              # endpad for Buffy the Vampire Slayer
#   validity Sat/19:30-05:00 endpad 20     # Saturday 7:30pm to Sunday 5am
proc readConfig { configFile } {
  global config selopt defaultConfigFile configMtime usedConfig tz cmdtz noargval

  array set alias {
    s         startpad
    e         endpad
    c         channel
    v         validity
    programme program
    p         program
    tz        timezone
    n         negpad
    m         maxneg
    f         forceneg
  }

  if {$configFile == ""} {
    # If no configFile was specified, then read default,
    # but don't choke if it isn't there.
    set configFile $defaultConfigFile
    if [catch {set f [open $configFile]}] return
  } else {
    set f [open $configFile]
  }

  set configMtime [file mtime "$configFile"]
  set usedConfig $configFile
  set linenum 0
  while {[gets $f line] >= 0} {
    incr linenum
    regsub "#.*" $line "" line
    if {[llength $line] == 0} continue
    catch {unset opt}
    set opt(linenum) $linenum
    set lline [llength $line]
    for {set i 0} {$i<$lline} {incr i} {
      set key [string tolower [lindex $line $i]]
      regsub "^-" $key "" key
      if [info exists alias($key)] {
        set key $alias($key)
      }
      if { $key == "timezone" } {
        incr i
        if { $cmdtz == 0 } {
          set tz [timezoneNumber [lindex $line $i]]
        }
      } elseif [info exists selopt($key)] {
        if [info exists noargval($key)] {
          set opt($key) $noargval($key)
        } else {
          incr i
          set opt($key) [lindex $line $i]
        }
      } else {
        puts "unrecognised keyword ($key) in file $configFile, line $linenum: $line"
        usage
        exit 0
      }
    }
    if [info exists opt(channel)] {
      set chans [split [string toupper $opt(channel)] ","]
      unset opt(channel)
    } else {
      set chans {""}
    }
    if [info exists opt(validity)] {
      set vl [validityNumbers $opt(validity)]
      if {$vl == ""} {
        puts "bad validity range ($opt(validity)) in file $configFile, line $linenum: $line"
        usage
        exit 0
      }
      set opt(validity) $vl
    }
    # Allow multiple config lines for the same channel.
    # Also used for defaults ($config()).
    foreach c $chans {
      lappend config($c) [array get opt]
    }
  }
  close $f
}

proc showConfig {} {
  global config noargval
  if {[array names config] < 2} return
  log " Sorted configuration options (last match used):-"
  # sort means we get config() first
  foreach chan [lsort [array names config]] {
    foreach c $config($chan) {
      catch { unset newopt }
      array set newopt $c
      catch {unset newopt(precedence)}
      catch {unset newopt(linenum)}
      set sel [expr {$chan == "" ? {} : "channel $chan"}]
      # Put these options first
      foreach n {program validity startpad endpad negpad} {
        if [info exists newopt($n)] {
          if {$n == "validity"} {
            lappend sel $n [validityString $newopt($n)]
          } else {
            lappend sel $n $newopt($n)
          }
          unset newopt($n)
        }
      }
      # show the rest
      foreach n [lsort [array names newopt]] {
        if {![info exists noargval($n)]} {
          lappend sel $n $newopt($n)
        } elseif {$newopt($n) == $noargval($n)} {
          lappend sel $n
        }
      }
      log "   $sel"
    }
  }
}

# Get config for program on $chan and starting at $time named $program.
# $name and $usekeys are just used for the log messages.
proc getConfig { {chan ""} {date ""} {time ""} {program ""} {name ""} {usekeys {}} } {
  global config tz

  set ucchan [string toupper $chan]
  if {![info exists config($ucchan)]} { return }
  foreach c $config($ucchan) {
    set sel [expr {$chan=="" ? {} : [list "is on $chan"]}]
    catch {unset newopt}
    array set newopt $c
    if [info exists newopt(validity)] {
      if {$time == ""} { continue }
      set s [lindex $newopt(validity) 0]
      set e [lindex $newopt(validity) 1]
      if {$s != "" && $e != ""} {
        if {[lindex $newopt(validity) 2]} {
          if {$date == ""} { continue }
          # get seconds since Sunday midnight
          set t [expr {[clock format [expr {86400*$date}] -format "%w"]*86400 + $time}]
        } else {
          set t $time
        }
        if {$s < $e ? ($t >= $s && $t < $e)
                    : ($t >= $s || $t < $e)} {
          lappend sel "starts in [validityString $newopt(validity)]"
        } else {
          continue
        }
      }
    }
    if [info exists newopt(program)] {
      if {$program == ""} { continue }
      if [ regexp -nocase -- $newopt(program) $program ] {
        lappend sel "matches '$newopt(program)'"
      } else {
        continue
      }
    }
    array set opt $c

    if [llength $sel] {
      set usevals {}
      foreach k $usekeys {
        if [info exists newopt($k)] {
          lappend usevals $k $newopt($k)
        }
      }
      if [llength $usevals] {
        log " $name ($program[expr {$date=="" || $time=="" ? "" : [clock format [expr 86400*$date+$time+$tz] -format " at %H:%M on %a"]}]) [join $sel " and "] so use $usevals"
      }
    }
  }

  return [array get opt]
}

proc setConfig { {cchan 0} {nchan 0} {cdate ""} {ndate ""} {ctime ""} {ntime ""} {ctitle ""} {ntitle ""} } {
  global startpadding endpadding suggestionRecQuality sugcancancelsug negativepadding forcenegpad

  array set copt [getConfig ""     $cdate $ctime $ctitle "Current recording" {endpad sugeq negpad forceneg}]
  array set copt [getConfig $cchan $cdate $ctime $ctitle "Current recording" {endpad sugeq negpad forceneg}]
  array set nopt [getConfig ""     $ndate $ntime $ntitle "Next recording"    {startpad sugqual maxneg}]
  array set nopt [getConfig $nchan $ndate $ntime $ntitle "Next recording"    {startpad sugqual maxneg}]

  set startpadding    [expr int($nopt(startpad) * 60)]
  set endpadding      [expr int($copt(endpad)   * 60)]
  set negativepadding [expr int($copt(negpad)   * 60)]
  set forcenegpad               $copt(forceneg)
  set suggestionRecQuality      $nopt(sugqual)
  set sugcancancelsug           $copt(sugeq)

  if { $negativepadding > $endpadding } {
    set negativepadding $endpadding
  }
  if { $nopt(maxneg) >= 0} {
    set maxnegpad     [expr int($nopt(maxneg)   * 60)]
    if {$maxnegpad < $negativepadding } {
      set negativepadding $maxnegpad
    }
  }
}

######################################################################
### Setup
######################################################################

# parse args
set runlikethewind false
set configFile ""
set usedConfig ""
set debug false
set checkInterval 60
set cmdtz 0
set debugfsid 0
set negfsid 0

# Save the command line options in cmdopt(X) for use by file reload, etc
for {set i 0} {$i < [llength $argv]} {incr i} {
  set arg [lindex $argv $i]
  switch -- $arg {
    -stop {if { [checkpid $pidfile]==0} {
        set pfchan [open $killfile "w"]
        puts $pfchan "Kill endpad.tcl, this file can be deleted"
        close $pfchan
        puts "Kill signal sent; may take up to a minute, you will not get an acknowledgement"
        puts "Check log with TiVoWeb to confirm exit."
      } else {
        puts stderr "No instance of EndPad running"
      }
      exit 0
    }
    -auto {file delete -force $pidfile}
    -run { set runlikethewind true}
    -debug {file delete -force $pidfile
      set runlikethewind true
      set checkInterval 5
      set debug true
    }
    -s -
    -startpad {incr i; set cmdopt(startpad)   [lindex $argv $i]}
    -e -
    -endpad   {incr i; set cmdopt(endpad)     [lindex $argv $i]}
    -n -
    -negpad   {incr i; set cmdopt(negpad)     [lindex $argv $i]}
    -f -
    -forceneg {incr i; set cmdopt(forceneg)   [lindex $argv $i]}
    -sugqual  {incr i; set cmdopt(sugqual)    [lindex $argv $i]}
    -sugeq    {        set cmdopt(sugeq)      0}
    -config   {incr i; set configFile         [lindex $argv $i]}
    -tz -
    -timezone {incr i; set tz [timezoneNumber [lindex $argv $i]]; set cmdtz 1}
    -fsid     {incr i; set debugfsid          [lindex $argv $i]}
    -negfsid  {incr i; set negfsid            [lindex $argv $i]}
    -? -
    -help     {usage; exit 0}
    default {puts "Argument $arg not recognised"
      usage
      exit 0
    }
  }
}

# Go load the config
reloadConfig $configFile

set nothingToDo true
foreach chan [array names config] {
  catch {unset opt}
  foreach optlist $config($chan) {
    array set opt $optlist
    if {([info exists opt(startpad)] && $opt(startpad) !=  0) ||
        ([info exists opt(endpad)]   && $opt(endpad)   !=  0) ||
        ([info exists opt(negpad)]   && $opt(negpad)   !=  0) ||
        ([info exists opt(sugqual)]  && $opt(sugqual)  != -1)} {
      set nothingToDo false
      break
    }
  }
}
if {$nothingToDo} {
  puts "No padding requested"
  if {$debug} {} else {
    exit 0
  }
}

# If endpad is already running, then warn and stop
if {[checkpid $pidfile] == 0} {
  puts stderr "EndPad is already running"
  exit 0
}

# If we get the -run flag then this must be the forked copy
# so create the pid file and get going
# else fork a copy in the background and quit
if { $runlikethewind } {
  set pfchan [open $pidfile "w"]
  puts $pfchan [pid]
  close $pfchan
} else {
  # fork off a copy of this script to be run in the background
  puts "EndPad v$epversion (c) 2005 Stuart Anderton"
  puts "Strictly no commercial distribution"
  puts "See file header for terms of use and distribution"
  puts "Switching to background..."
# puts "[info nameofexecutable] [info script] $argv -run >>& $errfile &"
  eval exec {[info nameofexecutable] [info script]} $argv {-run >>& $errfile &}
  exit 0
}

######################################################################
### The working code, only run in the forked copy
######################################################################

# Initialise stuff
set dbPoolSize [expr 100 * 1024]
set test 1
set datestamp "%A %H:%M:%S"
set lastrunrecid 0
set count 0
set newstarttime 0
set newstoptime 0
set cselectiontype 0
set endpadtoset 0
set gap 0
set cendtime 0
set cexistingendpadding 0
set cexistingstartpadding 0
set nexistingstartpadding 0
set seriestmsid {}
set episodetmsid {}
set errortime 0
set ccallsign 0
set ncallsign 0
set ctitle ""
set ntitle ""
set cstartdate ""
set nstartdate ""
set cstarttime ""
set nstarttime ""
catch {EnableTransactionHoldoff true}

# Open logfile

if {$debug} {
  set logchan "stdout"
  set logdate ""
} elseif {![file exists $logfile]} {
  set logchan [open $logfile "a"]
  set logdate [clock format [expr [file mtime "$logfile"]+$tz] -format "%j"]
} else {
  set logdate [clock format [expr [file mtime "$logfile"]+$tz] -format "%j"]
  set logchan [open $logfile "a"]
}

# Open summary file

if {![file exists $msgfile]} {
  set msgchan [open $msgfile "a"]
  set msgdate [clock format [expr [file mtime "$msgfile"]+$tz] -format "%U"]
} else {
  set msgdate [clock format [expr [file mtime "$msgfile"]+$tz] -format "%U"]
  set msgchan [open $msgfile "a"]
}

setConfig   ;# just so logheader is happy

# Check version
set db [dbopen $dbPoolSize]
RetryTransaction {
  set versionstring [dbobj [db $db open "/SwSystem/ACTIVE"] get Name]
}
set version [string range $versionstring 0 0]
dbclose $db

log " "
log " endpad.tcl $epversion log file, starting up"
logheader

if {$negfsid} {
  set db [dbopen $dbPoolSize]
  AddNegativeStartPadding $negfsid [expr int($cmdopt(negpad) * 60)]
  dbclose $db
  close $logchan
  close $msgchan
  file delete -force $pidfile
  file delete -force $killfile
  file delete -force $errfile
  exit 0
}

# Main loop continues forever unless kill signal recieved
while {$test} {

  # Error trap for those pesky db retry errors
  if {[catch {
      # Woken up, get info on current and next recordings
      log " Woken up"
      set db [dbopen $dbPoolSize]
      set isrec 0
      set dostartpadnow 0

      if {!$debugfsid} {
        # Read mfs for currently recording and next schedules programmes
        RetryTransaction {
          set currec [mfs scan "/Recording/Active" -start "3:" -count 1]
          set nextrec [mfs scan "/Recording/Active" -start "4:" -count 2]
        }
      } else {
        set found 0
        set currec ""
        set nextrec ""
        ForeachMfsFile fsid name type "/Recording/Active" "4:" {
          set nextbutonerec [list $fsid $name $type]
          if {[lindex [split [lindex $currec 1] ":"] 3] == $debugfsid} {
            set found 1
            break
          }
          set currec $nextrec
          set nextrec $nextbutonerec
        }
        if {!$found} {
          msg "Debug FSID $debugfsid not found - exit"
          dbclose $db
          close $logchan
          close $msgchan
          file delete -force $pidfile
          file delete -force $killfile
          file delete -force $errfile
          exit 0
        }
        set currec  [list $currec]
        set nextrec [list $nextrec $nextbutonerec]
      }
      scan [lindex [lindex $currec 0] 1] "%d:%d:%d:%d" rectype recstartdate recstarttime recfsid
      scan [lindex [lindex $nextrec 0] 1] "%d:%d:%d:%d" nextrectype nextrecstartdate nextrecstarttime nextrecfsid
      scan [lindex [lindex $nextrec 1] 1] "%d:%d:%d:%d" nextbutonerectype nextbutonerecstartdate nextbutonerecstarttime nextbutonerecfsid

      if {$rectype == 3 || $debugfsid} {

        # Recording active
        set isrec 1

        # Get details of currently recording programme
        RetryTransaction {
          set crec [db $db openid $recfsid]
          if {$version >= 3} {
            set crecbeh [dbobj $crec get RecordingBehavior]
            set cselectiontype [dbobj $crecbeh get PresentationBehavior]
          } else {
            set cselectiontype [dbobj $crec get SelectionType]
          }
          set cshowing [dbobj $crec get Showing]
          set cstation [dbobj $cshowing get Station]
          set ccallsign [dbobj $cstation get CallSign]
          set cprogram [dbobj $cshowing get Program]
          set ctitle [dbobj $cprogram get Title]
          set cstartdate [dbobj $cshowing get Date]
          set cstarttime [dbobj $cshowing get Time]
          set cduration [dbobj $cshowing get Duration]
          set cexistingendpadding [dbobj $crec get EndPadding]
          set cexistingstartpadding [dbobj $crec get StartPadding]
        }
        if {$cexistingstartpadding == ""} {set cexistingstartpadding 0}
        if {$cexistingendpadding == ""} {set cexistingendpadding 0}
        if {$cselectiontype == 6} {
          set ctypestring "(Suggestion)"
        } else {
          set ctypestring ""
        }

        # Time this recording ends
        set cendtime [expr $cstartdate  * 86400  + $cstarttime + $cduration + $cexistingendpadding]
        set cnewendtime $cendtime
      } else {
        set isrec 0
      }

      # Get details of next recording
      RetryTransaction {
        set nrec [db $db openid $nextrecfsid]
        if {$version >= 3} {
          set nrecbeh [dbobj $nrec get RecordingBehavior]
          set nselectiontype [dbobj $nrecbeh get PresentationBehavior]
        } else {
          set nselectiontype [dbobj $nrec get SelectionType]
        }
        set nshowing [dbobj $nrec get Showing]
        set nstation [dbobj $nshowing get Station]
        set ncallsign [dbobj $nstation get CallSign]
        set nprogram [dbobj $nshowing get Program]
        set nstartdate [dbobj $nshowing get Date]
        set nstarttime [dbobj $nshowing get Time]
        set nduration [dbobj $nshowing get Duration]
        set ntitle [dbobj $nprogram get Title]
        set nexistingstartpadding [dbobj $nrec get StartPadding]
      }
      if {$nexistingstartpadding == ""} {set nexistingstartpadding 0}
      # Time this recording starts
      set nextstart [expr $nstartdate  * 86400  + $nstarttime - $nexistingstartpadding]

      # Set max endpad to be the start of the next but one recording, to avoid an unforseen clash
      set longstop [expr $nextbutonerecstartdate  * 86400  + $nextbutonerecstarttime]
      if {$nselectiontype == 6} {
        set ntypestring "(Suggestion)"
        # Next program to record is a suggestion - check to see if we need to
        # delete it to make way for start padding on the following program.
        RetryTransaction {
          set nnrec [db $db openid $nextbutonerecfsid]
          if {$version >= 3} {
            set nnrecbeh [dbobj $nnrec get RecordingBehavior]
            set nnselectiontype [dbobj $nnrecbeh get PresentationBehavior]
          } else {
            set nnselectiontype [dbobj $nnrec get SelectionType]
          }
          set nnshowing [dbobj $nnrec get Showing]
          set nnstation [dbobj $nnshowing get Station]
          set nncallsign [dbobj $nnstation get CallSign]
          set nnprogram [dbobj $nnshowing get Program]
          set nnstartdate [dbobj $nnshowing get Date]
          set nnstarttime [dbobj $nnshowing get Time]
          set nnduration [dbobj $nnshowing get Duration]
          set nntitle [dbobj $nnprogram get Title]
          set nnexistingstartpadding [dbobj $nnrec get StartPadding]
        }
        # Close the database since we're done with it for a bit
        dbclose $db

        setConfig $ncallsign $nncallsign $nstartdate $nnstartdate $nstarttime $nnstarttime $ntitle $nntitle
        # We need to cancel the next suggestion if:
        #  The following program is not a suggestion and it's requested
        #   startpad overlaps the suggestion.
        set nextend [expr $nstartdate  * 86400  + $nstarttime + $nduration]
        if {$startpadding > 0 && $nnselectiontype != 6 &&
            $nextend > [expr $longstop - $startpadding]} {
          msg " Deleting suggestion $ntitle on $ncallsign to make room for $startpadding seconds startpad of $nntitle on $nncallsign"
          DeleteSuggestion $nextrecfsid
          # Do the startpadding of the program now to avoid another pesky
          # suggestion jumping in to our startpadding. Also do endpadding of
          # current even if it isn't time yet
          set dostartpadnow 1

          # Fix up the next recording information...
          set nselectiontype $nnselectiontype
          set ncallsign $nncallsign
          set nstartdate $nnstartdate
          set nstarttime $nnstarttime
          set nduration $nnduration
          set ntitle $nntitle
          set nexistingstartpadding $nnexistingstartpadding
          set nextrecfsid $nextbutonerecfsid
          set ntypestring ""
          if {$nexistingstartpadding == ""} {set nexistingstartpadding 0}
          set nextstart [expr $nstartdate  * 86400  + $nstarttime - $nexistingstartpadding]
        }
      } else {
        set ntypestring ""
        dbclose $db
      }

      if {$isrec == 1} {
        log " Recording $ctitle on $ccallsign $ctypestring now"
        log " Recording Scheduled End Time: [clock format [expr $cendtime+$tz] -format %H:%M]"
      } else {
        log " No recording in progress"
      }
      log " Next recording is $ntitle on $ncallsign $ntypestring at [clock format [expr $nextstart+$tz] -format %H:%M]"
      flush $logchan

      setConfig $ccallsign $ncallsign $cstartdate $nstartdate $cstarttime $nstarttime $ctitle $ntitle

      #Let's see if any end padding needs doing on the current recording
      if {$endpadding > 0 && $isrec == 1 && $endpadding > $cexistingendpadding} {
        if {$dostartpadnow != 0 || $debugfsid || [clock seconds] >= [expr $cendtime - 300] || [clock seconds] >= [expr $nextstart - $startpadding - 60]} {

          # Check if end padding can be added
          set gap [expr $nextstart - $cendtime + $cexistingendpadding]
          log " Gap to next programme is $gap seconds"

          #### Check if negative padding needs to be set

          if { $negativepadding > 0 && $gap < $negativepadding && ([info exists lastfsidnegpad] == 0 || $lastfsidnegpad != $nextrecfsid) && ($ccallsign != $ncallsign || $forcenegpad != 0) && !($cselectiontype == 6 && $nselectiontype != 6)} {
            set negativepaddingtoadd [expr $negativepadding - $gap]
            if { $negativepaddingtoadd > $negativepadding } {
              set negativepaddingtoadd $negativepadding
            }
            if { $negativepaddingtoadd > 0 } {
              log " Need to add $negativepaddingtoadd seconds of negative padding to $ntitle"
              set newnextrecfsid [AddNegativeStartPadding $nextrecfsid $negativepaddingtoadd]

              if { $newnextrecfsid == -1 } {
                log " Negative padding failed"
              } else {
                set nextrecfsid $newnextrecfsid
                set db [dbopen $dbPoolSize]
                RetryTransaction {
                  set nrec [db $db openid $nextrecfsid]
                  set nshowing [dbobj $nrec get Showing]
                  set nstartdate [dbobj $nshowing get Date]
                  set nstarttime [dbobj $nshowing get Time]
                  set nduration [dbobj $nshowing get Duration]
                  set nexistingstartpadding [dbobj $nrec get StartPadding]
                }
                if {$nexistingstartpadding == ""} {set nexistingstartpadding 0}
                dbclose $db
                # Time this recording starts
                set nextstart [expr $nstartdate  * 86400  + $nstarttime - $nexistingstartpadding]
                set gap [expr $nextstart - $cendtime + $cexistingendpadding]
                log " Gap to next programme is now $gap seconds"
                set lastfsidnegpad $nextrecfsid
              }
            }
            flush $logchan
          }

          if {$gap < $endpadding && $nselectiontype == 6 && (($ncallsign != $ccallsign && $sugcancancelsug == 1) || $cselectiontype != 6)} {
            DeleteSuggestion $nextrecfsid
            msg " Deleted the suggestion $ntitle on $ncallsign from To Do List to enable padding of $ctitle on $ccallsign $ctypestring"
            set endpadtoset $endpadding
          } else {
            # If current is a suggestion, don't step on a real recording's
            # startpad.
            if {$nselectiontype != 6 && $cselectiontype == 6} {
              set gap [expr $gap - $startpadding]
            }
            if {$gap > 2} {
              if {$gap >= $endpadding} {
                set endpadtoset $endpadding
              } else {
                set endpadtoset $gap
              }
            } else {
              set endpadtoset 0
            }
          }

          # check padding will not go past longstop
          if {[expr $cendtime - $cexistingendpadding + $endpadtoset] >= [expr $longstop - 30]} {
            set endpadtoset [expr $longstop - $cendtime  + $cexistingendpadding  - 30]
          }

          if {$endpadtoset > 0} {
            RetryTransaction {
              # Alter the database record to add the padding
              set db [dbopen $dbPoolSize]
              set rec [db $db openid $recfsid]
              dbobj $rec set EndPadding $endpadtoset
              dbclose $db
            }
            msg " Added $endpadtoset seconds end padding to $ctitle on $ccallsign $ctypestring"
            set cnewendtime [expr $cendtime - $cexistingendpadding + $endpadtoset]

          } else {
            log " Can't add end padding to $ctitle"
          }
        } else {
          log " Not time to add end padding yet"
        }

      } else {
        log " No end padding required:"
        if {$endpadding <= $cexistingendpadding} {
          log " End padding of $cexistingendpadding seconds already set"
        }
        if {$endpadding == 0} {
          log " No end padding requested"
        }
        if {$isrec != 1} {
          log " No recording in progress"
        }
      }

      # Set time to wake up for next end padding job
      if {[clock seconds] < [expr $cendtime - 300]} {
        set nextepwake [expr $cendtime - 300]
      } else {
        if {$nduration < 400} {
          set nextepwake [expr $nextstart + 60]
        } else {
          set nextepwake [expr $nextstart + $nduration - 300]
        }
      }
      log " Next end padding setting due [clock format [expr $nextepwake+$tz] -format %H:%M]"


      # End padding sorted, so let's see if we need to add start padding to the next recording
      if {$startpadding != 0 && $startpadding > $nexistingstartpadding} {

        # Are we at the last possible time to add the padding, or do we want to get in before another suggestion does
        if {$dostartpadnow != 0 || $debugfsid || $nextstart - [clock seconds] <= [expr $startpadding + 60]} {

          # If so then try to add start padding
          # if there's a "real" recoding going on we need to see if there is space available
          if {$isrec == 1} {

            # Check if padding can be added
            set gap [expr $nextstart - $cnewendtime]
            log " Gap to previous recording is $gap seconds"
            if {$gap >= $startpadding} {
              set startpadtoset $startpadding
            } else {
              set startpadtoset $gap
            }
          } else {
            set startpadtoset $startpadding
          }

          if {$startpadtoset > 0} {
            RetryTransaction {
              # Alter the database record to add the padding
              set db [dbopen $dbPoolSize]
              set rec [db $db openid $nextrecfsid]
              dbobj $rec set StartPadding $startpadtoset
                  # Add Bookmark for this showing to start at the original time
                  set nbm [db $db createsub Bookmark $rec]
                  dbobj $nbm set TimeMs [expr $startpadtoset * 1000]
                  dbobj $rec set Bookmark $nbm
							dbclose $db
            }
            msg " Added $startpadtoset seconds start padding to $ntitle on $ncallsign $ntypestring"
          } else {
            log " Can't add start padding to $ntitle"
          }

          # Change recording quality of suggestion if requested
          if { $nselectiontype == 6 && $suggestionRecQuality >= 0 } {
            RetryTransaction {
              set db [dbopen $dbPoolSize]
              set rec [db $db openid $nextrecfsid]
              dbobj $rec set RecordQuality $suggestionRecQuality
              dbclose $db
            }
            msg " Suggestion Recording Quality set to $suggestionRecQuality for $ntitle on $ncallsign"
          }

          set nextspwake $nextepwake
        } else {

          # Else set wake up for before next programme starts
          log " Not time to add start padding yet"
          set nextspwake [expr $nextstart - [expr $startpadding + 60]]
        }
      } else {
        log " Can't add start padding to $ntitle"
        if {$startpadding == 0} {
          log " No start padding requested"
        }
        if {$startpadding <= $nexistingstartpadding} {
          log " Start padding of $nexistingstartpadding seconds already set"
        }
        set nextspwake [expr $nextstart + 600]
      }
      log " Next start padding setting due [clock format [expr $nextspwake+$tz] -format %H:%M]"

      if {$debugfsid} {
        msg "Exit after performing padding for programme $debugfsid"
        close $logchan
        close $msgchan
        file delete -force $pidfile
        file delete -force $killfile
        file delete -force $errfile
        exit 0
      }

      # Which wake up comes first?
      if {$nextspwake <= $nextepwake} {
        set nexttime $nextspwake
      } else {
        set nexttime $nextepwake
      }

      # Max 30 mins between wake ups to catch new recordings
      if {[expr $nexttime - [clock seconds]] > 1800} {
        set nexttime [expr [clock seconds] + 1800]
      }
      if {$nexttime < [clock seconds]} {
        set nexttime [expr [clock seconds] + 30]
      }
      log " Next wake up will be [clock format [expr $nexttime+$tz] -format %H:%M]"

      # Zero variables in case of old numbers getting used on next loop - so many nexted ifs that not all
      # logic is re-run each time
      set nrec 0
      set nselectiontype 0
      set nshowing 0
      set nduration 0
      set nstation 0
      set ncallsign 0
      set nprogram 0
      set ntitle 0
      set nexistingstartpadding 0
      set crec 0
      set cselectiontype 0
      set cshowing 0
      set cstation 0
      set ccallsign 0
      set cprogram 0
      set ctitle 0
      set cduration 0
      set cexistingendpadding 0
      set cexistingstartpadding 0
      set endpadtoset 0
      set startpadtoset 0
      set cnewendtime 0
      set cstartdate ""
      set nstartdate ""
      set cstarttime ""
      set nstarttime ""

      # switch log files at midnight
      if {$logdate != "" && $logdate != [clock format [expr [clock seconds]+$tz] -format "%j"] } {
        flush $logchan
        close $logchan
        catch {file copy -force $logfile $Ologfile}
        catch {file delete -force $logfile}
        set logchan [open $logfile "a"]
        log " endpad.tcl $epversion log file, log rotated"
        logheader
        set logdate [clock format [expr [clock seconds]+$tz] -format "%j"]
      }

      # switch summary files every Sunday midnight
      if {$msgdate != "" && $msgdate != [clock format [expr [clock seconds]+$tz] -format "%U"] } {
        flush $msgchan
        close $msgchan
        catch {file copy -force $msgfile $Omsgfile}
        catch {file delete -force $msgfile}
        set msgchan [open $msgfile "a"]
        log " summary file $msgfile rotated"
        set msgdate [clock format [expr [clock seconds]+$tz] -format "%U"]
      }

      # Wait until next wake up, checking every 60s for a kill message
      puts -nonewline $logchan "Sleeping, checking every minute for kill signal."
      set wait 1
      while {$wait} {
        puts -nonewline $logchan "."
        flush $logchan
        set togo [expr $nexttime - [clock seconds]]
        if {$togo <= $checkInterval} {
          after [expr $togo * 1000]
        } else {
          after [expr $checkInterval * 1000]
        }
        if {[clock seconds] >= $nexttime} {
          set wait 0
        }
        if { [file exists $killfile] } {
          puts "EndPad has stopped"
          puts $logchan " "
          log " Stopped due to kill signal"
          file delete -force $pidfile
          file delete -force $killfile
          file delete -force $errfile
          exit 0
        }
        if { $configMtime > 0 && [file mtime "$usedConfig"] != $configMtime } {
          puts $logchan " "
          log "Reloading config file due to file change"
          reloadConfig $configFile
          setConfig   ;# just so logheader is happy
          logheader
          set wait 0
        }
      }
      puts $logchan " "
      flush $logchan
    } result]} {

    # Runs if error detected
    global errorInfo
    log " Error detected"
    log $errorInfo
    log " Attempting to recover"
    set lasterror $errortime
    set errortime [clock seconds]
    if {$debug || [expr $errortime - $lasterror] < 120} {
      # Two errors in 2 mins? Bye bye!
      log " Too many errors, stopping"
      close $logchan
      close $msgchan
      exit
    }

    # wait a bit for the db to get clear, then off we go again
    after 30000
    set nrec 0
    set nselectiontype 0
    set nshowing 0
    set nduration 0
    set nshowingfsid 0
    set nstation 0
    set nstationfsid 0
    set ncallsign 0
    set nprogram 0
    set ntitle 0
    set nexistingstartpadding 0
    set crec 0
    set cselectiontype 0
    set cshowing 0
    set cshowingfsid 0
    set cstation 0
    set cstationfsid 0
    set ccallsign 0
    set cprogram 0
    set ctitle 0
    set cduration 0
    set cexistingendpadding 0
    set cexistingstartpadding 0
    set endpadtoset 0
    set startpadtoset 0
    set cnewendtime 0
    set cstartdate ""
    set nstartdate ""
    set cstarttime ""
    set nstarttime ""
  }
}

close $logchan
close $msgchan
